import unittest
from internals import Internals
from problems import UnresolvedProblems
from mapping import Mappings
from monster_place import MonsterPlaceTests
from monster_combine import MonsterCombineTests
from MPro_combine import VictorCombineTests
from MPro_place import MProPlaceTester
from multivictor import MultivictorPlaceTests
from walton import WaltonTests

if __name__ == '__main__':
    unittest.main()



########################################################################################################################

__doc__ = \
    """
Place followup
    """

########################################################################################################################

from typing import Optional, Dict, List
from warnings import warn

from rdkit import Chem

from ._place_modes import _MonsterBlend

class _MonsterPlace(_MonsterBlend):
    """
    * _MonsterBase
    * _MonsterCommunal
    * _MonsterTracker
    * _MonsterMerge
    * _MonsterCombine
    * _MonsterMap
    * _MonsterChimera
    * _MonsterRefine
    * (_MonsterFull, _MonsterPartial, _MonsterNone) >
    * _MonsterBlend
    * _MonsterPlace
    """

    def place(self,
              mol: Chem.Mol,
              attachment: Optional[Chem.Mol] = None,
              custom_map: Optional[Dict[str, Dict[int, int]]] = None,
              merging_mode: str = 'expansion'):
        """
        Positioned a given mol based on the hits. (Main entrypoint)
        accepts the argument `merging_mode`, by default it is "expansion",
        but was "permissive_none",
        which call `.by_expansion` and `.no_blending(broad=True)` respectively.
        "off" (does nothing except fill the attribute ``initial_mol``),
        "full" (`.full_blending()`),
        "partial" (`.partial_blending()`)
        and "none" (`.no_blending()`, but less thorough)
        are accepted.

        :param mol:
        :param attachment:
        :param custom_map:
        :param merging_mode:
        :return:
        """
        if not mol.HasProp('_Name'):
            mol.SetProp('_Name', 'followup')
        self.initial_mol, self.attachment = self._parse_mol_for_place(mol, attachment)
        if custom_map:
            self.custom_map: Dict[str, Dict[int, int]] = custom_map
        # do the mol names match?
        self._validate_custom_map()
        # Reset
        self.unmatched = []
        self.mol_options = []
        # do calculations
        self.merging_mode = merging_mode
        if merging_mode == 'off':
            pass
        elif merging_mode == 'full':
            self.full_blending()
        elif merging_mode == 'partial':
            self.partial_blending()
        elif merging_mode == 'none_permissive' or merging_mode == 'permissive_none':
            self.no_blending(broad=True)
        elif merging_mode == 'none':
            self.no_blending()
        elif merging_mode == 'expansion':
            self.by_expansion()
        else:
            valid_modes = ('full', 'partial', 'none', 'none_permissive', 'off', 'expansion')
            raise ValueError(
                f"Merging mode can only be {'| '.join(valid_modes)}, not '{merging_mode}'")
        return self

    def place_smiles(self,
                     smiles: str,
                     long_name: Optional[str]=None,
                     **kwargs):
        mol = Chem.MolFromSmiles(smiles)
        mol.SetProp('_Name', long_name if long_name else 'followup')
        self.place(mol=mol, **kwargs)
        return self

    def _parse_mol_for_place(self,
                             mol: Chem.Mol,
                             attachment: Optional[Chem.Mol] = None):
        # ------------- store mol ---------------------------------------
        if mol.HasSubstructMatch(self.dummy) and attachment:
            pass
        elif mol.HasSubstructMatch(self.dummy):
            warn('No attachment atom provided but dummy atom present --- ignoring.')
            attachment = None
        elif attachment:
            warn('Attachment atom provided but dummy atom not present --- ignoring.')
            attachment = None
        else:
            attachment = None
        return mol, attachment

    def _validate_custom_map(self):
        """
        What on earth has the user submitted as custom_map?

        :return:
        """
        # add missing keys
        hit_names:List[str] = []
        for hit in self.hits: #: Chem.Mol
            hit_name:str = hit.GetProp('_Name')
            hit_names.append(hit_name)
            if hit_name not in self.custom_map:
                self.custom_map[hit_name] = {}
        # check for incorrect keys
        for hit_name in self.custom_map:
            if hit_name not in hit_names:
                raise ValueError(f"Custom map contains key '{hit_name}' which is not in hits ({hit_names}).")




from IPython.display import SVG, display, HTML
import nglview as nv
from rdkit import Chem
from rdkit.Chem import Draw, AllChem
from matplotlib.colors import ColorConverter
from ..branding import divergent_colors
from functools import singledispatchmethod
from typing import Tuple, List, Dict
from io import StringIO
import re


def draw(mol, color_map, x=200, y=200):
    d = Draw.rdMolDraw2D.MolDraw2DSVG(x, y)
    flat = Chem.Mol(mol)
    AllChem.Compute2DCoords(flat)
    color_map2 = {i: ColorConverter().to_rgb(n) for i, n in color_map.items()}
    Draw.rdMolDraw2D.PrepareAndDrawMolecule(d,
                                            flat,
                                            highlightAtoms=list(color_map.keys()),
                                            highlightAtomColors=color_map2,
                                            )
    d.FinishDrawing()
    display(SVG(d.GetDrawingText()))


def get_idx(name, origin, default=None):
    rex = re.match(name + '\.(\d+)', origin)
    if rex:
        return int(rex.group(1))
    return default

class _MonsterUtilCompare:

    def show_comparison(self):
        """"
        Show the atom provenance of the follow molecule.

        Parameters:

        * nothing -> uses self.origin_from_mol
        * hit2followup -> uses the hit2followup dict
        """
        # ------- followup -----------------------
        #  origins is like [['MOL.7'], ['MOL.8'], ['MOL.9'], [], ...
        origins = self.origin_from_mol(self.positioned_mol)
        n_colors = sum(map(bool, origins))  # bool([]) is sensibly False in Python
        colors = divergent_colors[n_colors]
        mapped_idxs = [i for i, m in enumerate(origins) if m]
        # d = Draw.rdMolDraw2D.MolDraw2DCairo(500, 500)
        followup_color_map = dict(zip(mapped_idxs, colors))
        # later: draw(victor.minimized_mol, color_map)

        # -------------- hits ------------------
        for mol in self.hits:
            name = mol.GetProp('_Name')
            # convert ['MOL.7'] to a list of indices of MOL
            matched = [[get_idx(name, o) for o in ori] for ori in origins]
            # make a color_map
            color_map = {}
            for followup_idx, target_idxs in enumerate(matched):
                for i in target_idxs:  # there ought to only zero or one...
                    if i is None:
                        continue
                    color_map[i] = followup_color_map[followup_idx]
            print(f'hit {name}')  # legit print, not a debug scar
            draw(mol, color_map, 300, 300)
        print('Followup') # legit print, not a debug scar
        draw(self.positioned_mol, followup_color_map, 300, 300)


    def convert_origins_to_custom_map(self) -> Dict[str, Dict[int, int]]:
        """
        The origins stored in the followup differ in format from the custom_map.
        The former is a list of lists of hit_name+atom_index,
        while the latter is a dictionary of hit_name to dictionary of
        hit atom indices to _intended_ followup index.
        This method converts the former to the latter.

        :return:
        """
        custom_map: Dict[str, Dict[int, int]] = {}
        # `origins_from_mol` uses self.positioned_mol by default... what if it is not set?
        origins:List[List[str]] = self.origin_from_mol()  # noqa It is in utility
        for hit in self.hits:  #:Chem.Mol
            name:str = hit.GetProp('_Name')
            # "default=-1-fi" is to assign a unique negative number to prevent anything mapping to the target index.
            custom_map[name] = {get_idx(name, o, default=-1-fi): fi for fi, ori in enumerate(origins) for o in ori}
        return custom_map

    def _to_nglview_and_legend(self, show_positioned_mol:False) -> Tuple[nv.NGLWidget, str]:
        """
        This is called by both ``Monster.to_nglview`` and ``Victor.to_nglview``
        :return:
        """
        color_series = iter(divergent_colors[len(self.hits)])
        legend = ''
        view = nv.NGLWidget()
        for mol in self.hits:
            colorValue = next(color_series) if not mol.HasProp('_color') else mol.GetProp('_color')
            self.add_mol_to_nglview(colorValue=colorValue, mol=mol, view=view)

            legend += f'<span style="color: {colorValue}">{mol.GetProp("_Name")}</span> '
        if show_positioned_mol and self.positioned_mol:
            self.add_mol_to_nglview(colorValue='white', mol=self.positioned_mol, view=view)
            legend += f'<span>positioned followup (in white)</span> '

        return view, legend

    def add_mol_to_nglview(self, view: nv.NGLWidget, mol: Chem.Mol, colorValue: str) -> nv.component.ComponentViewer:
        if not mol:
            raise ValueError(
                'One of the hits is None: if user manual tinkering happened, please run monster.fix_hits')
        fh = StringIO(Chem.MolToMolBlock(mol))  # I want atom names
        comp: nv.component.ComponentViewer = view.add_component(fh,  # noqa it's there.
                                                                name=mol.GetProp('_Name'),
                                                                ext='mol')
        # _color business stems from Walton.
        comp.update_ball_and_stick(colorValue=colorValue, multipleBond=True)
        return comp

    def to_nglview(self, print_legend: bool = False) -> nv.NGLWidget:
        """
        This is not the same method as in Victor.
        generates a NGLWidget (``IPython.display.display`` will show it)
        with the compounds and the merged if present.
        To override the colours:
        The colours will be those in the ``Chem.Mol``'s property ``_color`` if present.

        Returns -> nv.NGLWidget
        """
        view, legend = self._to_nglview_and_legend(show_positioned_mol=True)
        if print_legend:
            display(HTML(legend))
        # async madness: disabled for now.
        # view.center(f'[{self.ligand_resn}]')
        return view

